from aimu.tools.client import MCPClient
from aimu import paths

from fastmcp import FastMCP

MCP_SERVERS_FILE = str(paths.package / "tools" / "servers.py")


def test_mcp_client_with_config():
    config = {
        "mcpServers": {"echo": {"command": "python", "args": [MCP_SERVERS_FILE]}},
    }

    client = MCPClient(config=config)
    tools = client.list_tools()

    assert isinstance(tools, list)
    assert len(tools) > 0


def test_mcp_client_with_file():
    client = MCPClient(file=MCP_SERVERS_FILE)
    tools = client.list_tools()

    assert isinstance(tools, list)
    assert len(tools) > 0


def test_mcp_client_with_server():
    mcp = FastMCP("TestTools")

    @mcp.tool()
    def echo(echo_string: str) -> str:
        return echo_string

    client = MCPClient(server=mcp)
    tools = client.list_tools()

    assert isinstance(tools, list)
    assert len(tools) > 0

    response = client.call_tool("echo", {"echo_string": "test"})

    assert response.content[0].text == "test"

    # call the tools a second time to ensure the server is running
    response = client.call_tool("echo", {"echo_string": "test again"})

    assert response.content[0].text == "test again"


def test_none_type_tool_response():
    mcp = FastMCP("TestTools")

    @mcp.tool()
    def none_response() -> None:
        return None

    client = MCPClient(server=mcp)
    tools = client.list_tools()

    assert isinstance(tools, list)
    assert len(tools) > 0

    response = client.call_tool("none_response", {})

    assert len(response.content) == 0


def test_multiple_tool_calls_client_reused():
    mcp = FastMCP("TestTools")

    @mcp.tool()
    def none_response() -> None:
        return None

    client = MCPClient(server=mcp)
    tools = client.list_tools()

    assert isinstance(tools, list)
    assert len(tools) > 0

    client.call_tool("none_response", {})
    mcp_client = client.client

    client.call_tool("none_response", {})

    assert mcp_client == client.client



from typing import List, Dict, Any
from tinydb import TinyDB
from datetime import datetime


class ConversationManager:
    def __init__(self, db_path: str = "chat_history.json", use_last_conversation: bool = False):
        self.__db_path = db_path

        self.__db = TinyDB(db_path)
        self.__conversations_table = self.__db.table("conversations")

        all = self.__conversations_table.all()
        if use_last_conversation and len(all) > 0:
            last = all[-1]
            doc_id = last.doc_id
            messages = self.__conversations_table.get(doc_id=last.doc_id)["messages"]
        else:
            doc_id, messages = self.create_new_conversation()

        self.__doc_id = doc_id
        self.__messages = messages

    def __del__(self):
        self.__db.close()

    def create_new_conversation(self) -> tuple[str, list]:
        messages = []
        return self.__conversations_table.insert({"messages": messages}), messages

    @property
    def messages(self) -> List[Dict[str, Any]]:
        return self.__messages.copy()

    def update_conversation(self, messages: List[Dict[str, Any]]) -> None:
        # Add a timestamp to new messages
        timestamp = datetime.now().isoformat()

        for message in messages[len(self.messages) :]:
            message["timestamp"] = timestamp
            self.__messages.append(message)

        self.__conversations_table.update({"messages": self.__messages}, doc_ids=[self.__doc_id])

    def close(self) -> None:
        self.__db.close()



from ..base_client import Model, ModelClient

import ollama
import logging
from typing import Iterator, Optional

logger = logging.getLogger(__name__)


class OllamaModel(Model):
    GPT_OSS_20B = "gpt-oss:20b"

    LLAMA_3_1_8B = "llama3.1:8b"
    LLAMA_3_2_3B = "llama3.2:3b"
    # LLAMA_3_3_70B = "llama3.3:70b"

    GEMMA_3_12B = "gemma3:12b"

    PHI_4_14B = "phi4:14b"
    PHI_4_MINI_3_8B = "phi4-mini:3.8b"

    DEEPSEEK_R1_8B = "deepseek-r1:8b"

    MISTRAL_7B = "mistral:7b"
    MISTRAL_NEMO_12B = "mistral-nemo:12b"
    MISTRAL_SMALL_3_2_24B = "mistral-small3.2:24b"

    QWEN_3_8B = "qwen3:8b"

    SMOLLM2_1_7B = "smollm2:latest"  # "smollm2:1.7b" error downloading model, using latest for now


class OllamaClient(ModelClient):
    MODELS = OllamaModel

    TOOL_MODELS = [
        MODELS.GPT_OSS_20B,
        MODELS.QWEN_3_8B,
        # MODELS.MISTRAL_7B, # issue with tool usage
        # MODELS.MISTRAL_NEMO_12B, # issue with tool usage
        MODELS.MISTRAL_SMALL_3_2_24B,
        MODELS.LLAMA_3_1_8B,
        MODELS.LLAMA_3_2_3B,
        # MODELS.LLAMA_3_3_70B,
        MODELS.SMOLLM2_1_7B,
    ]

    THINKING_MODELS = [
        MODELS.GPT_OSS_20B,
        MODELS.DEEPSEEK_R1_8B,
        MODELS.QWEN_3_8B,
    ]

    def __init__(self, model: OllamaModel, system_message: Optional[str] = None, model_keep_alive_seconds: int = 60):
        super().__init__(model, None, system_message)

        # TODO extend model_keep_alive_seconds to other model clients
        self.model_keep_alive_seconds = model_keep_alive_seconds

        self.thinking = True if model in self.THINKING_MODELS else False

        ollama.pull(model.value)

    def _update_generate_kwargs(self, generate_kwargs: Optional[dict] = None) -> dict[str, str]:
        if not generate_kwargs:
            generate_kwargs = self.default_generate_kwargs

        if "max_tokens" in generate_kwargs:
            generate_kwargs["num_predict"] = generate_kwargs.pop("max_tokens")

        return generate_kwargs

    def generate(self, prompt: str, generate_kwargs: Optional[dict] = None) -> str:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        response = ollama.generate(
            model=self.model.value,
            prompt=prompt,
            options=generate_kwargs,
            think=self.thinking,
            keep_alive=self.model_keep_alive_seconds,
        )

        return response["response"] if not self.thinking else response.response

    def generate_streamed(self, prompt: str, generate_kwargs: Optional[dict] = None) -> Iterator[str]:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        response = ollama.generate(
            model=self.model.value,
            prompt=prompt,
            options=generate_kwargs,
            stream=True,
            think=self.thinking,
            keep_alive=self.model_keep_alive_seconds,
        )

        for response_part in response:
            yield response_part["response"]

    def chat(self, user_message: str, generate_kwargs: Optional[dict] = None, use_tools: bool = True) -> str:
        generate_kwargs, tools = self._chat_setup(user_message, generate_kwargs, use_tools)

        response = ollama.chat(
            model=self.model.value,
            messages=self.messages,
            options=generate_kwargs,
            tools=tools,
            think=self.thinking,
            keep_alive=self.model_keep_alive_seconds,
        )

        if response["message"].tool_calls:
            tool_calls = []
            for tool_call in response["message"].tool_calls:
                tool_calls.append(
                    {
                        "name": tool_call.function.name,
                        "arguments": tool_call.function.arguments,
                    }
                )

            self._handle_tool_calls(tool_calls, tools)

            if response["message"].thinking:
                self.messages[-2]["thinking"] = response["message"].thinking

            response = ollama.chat(
                model=self.model.value,
                messages=self.messages,
                options=generate_kwargs,
                tools=tools,
                think=self.thinking,
                keep_alive=self.model_keep_alive_seconds,
            )

        self.messages.append({"role": response["message"].role, "content": response["message"].content})

        if response["message"].thinking:
            self.messages[-1]["thinking"] = response["message"].thinking

        return response["message"].content

    def chat_streamed(
        self, user_message: str, generate_kwargs: Optional[dict] = None, use_tools: bool = True
    ) -> Iterator[str]:
        generate_kwargs, tools = self._chat_setup(user_message, generate_kwargs, use_tools)

        response = ollama.chat(
            model=self.model.value,
            messages=self.messages,
            options=generate_kwargs,
            tools=tools,
            stream=True,
            think=self.thinking,
            keep_alive=self.model_keep_alive_seconds,
        )

        response_part = next(response)

        # If the model is thinking, we need to capture the thinking before processing tools and streaming the response.
        thinking = ""
        if response_part["message"].thinking:
            thinking = response_part["message"].thinking
            for response_part in response:
                if response_part["message"].thinking:
                    thinking += response_part["message"].thinking
                else:
                    break

        if response_part["message"].tool_calls:
            tool_calls = []
            for tool_call in response_part["message"].tool_calls:
                tool_calls.append(
                    {
                        "name": tool_call.function.name,
                        "arguments": tool_call.function.arguments,
                    }
                )

            self._handle_tool_calls(tool_calls, tools)

            if thinking:
                self.messages[-2]["thinking"] = thinking  # TODO: correct functionality if there are multiple tool calls
                thinking = ""

            response = ollama.chat(
                model=self.model.value,
                messages=self.messages,
                options=generate_kwargs,
                tools=tools,
                stream=True,
                think=self.thinking,
                keep_alive=self.model_keep_alive_seconds,
            )

            response_part = next(response)

            # For the response after the tool call, we need to capture the thinking again if it exists.
            thinking = ""
            if response_part["message"].thinking:
                thinking = response_part["message"].thinking
                for response_part in response:
                    if response_part["message"].thinking:
                        thinking += response_part["message"].thinking
                    else:
                        break

        content = response_part["message"].content
        yield content

        for response_part in response:
            content += response_part["message"].content
            yield response_part["message"].content

        message = {"role": response_part["message"].role, "content": content}

        if thinking:
            message["thinking"] = thinking

        self.messages.append(message)



from pathlib import Path

root = Path(__file__).parents[1]

# Path to the tests directory
tests = root / "tests"

# Path to the aimu package directory
package = root / "aimu"

# Path to transient output directory
output = root / "output"



"""
Tests for ConversationManager class.
"""

import pytest
import os
import tempfile
from tinydb import TinyDB

from aimu.memory import ConversationManager


@pytest.fixture
def temp_db_path():
    """Create a temporary database file for testing."""

    with tempfile.NamedTemporaryFile(suffix=".json", delete=False) as f:
        db_path = f.name

    yield db_path

    if os.path.exists(db_path):
        os.unlink(db_path)


def test_init_new_conversation(temp_db_path):
    """Test initialization with new conversation."""

    manager = ConversationManager(db_path=temp_db_path, use_last_conversation=False)

    assert manager.messages == []

    manager.close()


def test_init_use_last_conversation_empty_db(temp_db_path):
    """Test initialization with use_last_conversation=True on empty database."""

    manager = ConversationManager(db_path=temp_db_path, use_last_conversation=True)

    assert manager.messages == []

    manager.close()


def test_init_use_last_conversation_existing_db(temp_db_path):
    """Test initialization with use_last_conversation=True on existing database."""

    # First, create a conversation in the main table to simulate existing data
    db = TinyDB(temp_db_path)
    conversations_table = db.table("conversations")
    doc_id = conversations_table.insert(
        {"messages": [{"role": "user", "content": "Hello", "timestamp": "2023-01-01T00:00:00"}]}
    )

    # Also add to main table for the use_last_conversation logic to find it
    db.insert({"doc_id": doc_id})
    db.close()

    # Now test loading the last conversation
    manager = ConversationManager(db_path=temp_db_path, use_last_conversation=True)

    assert len(manager.messages) == 1
    assert manager.messages[0]["role"] == "user"
    assert manager.messages[0]["content"] == "Hello"
    assert "timestamp" in manager.messages[0]

    manager.close()


def test_create_new_conversation(temp_db_path):
    """Test _create_new_conversation method."""

    manager = ConversationManager(db_path=temp_db_path)

    doc_id, messages = manager.create_new_conversation()

    assert isinstance(doc_id, int)
    assert messages == []

    manager.close()


def test_update_conversation(temp_db_path):
    """Test update_conversation method."""

    manager = ConversationManager(db_path=temp_db_path)

    test_messages = [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi there!"}]

    manager.update_conversation(test_messages)

    # Verify the messages were stored with timestamps
    updated_messages = manager.messages
    assert len(updated_messages) == 2
    assert updated_messages[0]["role"] == "user"
    assert updated_messages[0]["content"] == "Hello"
    assert "timestamp" in updated_messages[0]
    assert updated_messages[1]["role"] == "assistant"
    assert updated_messages[1]["content"] == "Hi there!"
    assert "timestamp" in updated_messages[1]
    # Both messages should have the same timestamp
    assert updated_messages[0]["timestamp"] == updated_messages[1]["timestamp"]

    manager.close()


def test_update_conversation_multiple_times(temp_db_path):
    """Test updating conversation multiple times."""

    manager = ConversationManager(db_path=temp_db_path)

    # First update
    messages1 = [{"role": "user", "content": "Hello"}]
    manager.update_conversation(messages1)

    # Second update - only adding new messages
    messages2 = [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi there!"}]
    manager.update_conversation(messages2)

    # Verify that only the new message was appended
    stored_messages = manager.messages
    assert len(stored_messages) == 2
    assert stored_messages[0]["role"] == "user"
    assert stored_messages[0]["content"] == "Hello"
    assert "timestamp" in stored_messages[0]
    assert stored_messages[1]["role"] == "assistant"
    assert stored_messages[1]["content"] == "Hi there!"
    assert "timestamp" in stored_messages[1]

    manager.close()


def test_default_db_path():
    """Test default database path."""

    manager = ConversationManager()

    assert os.path.exists("chat_history.json")

    manager.close()

    # Clean up default file
    os.unlink("chat_history.json")


def test_multiple_conversations_in_same_db(temp_db_path):
    """Test multiple conversations can be stored in the same database."""

    # Create first conversation
    manager1 = ConversationManager(db_path=temp_db_path, use_last_conversation=False)
    messages1 = [{"role": "user", "content": "First conversation"}]
    manager1.update_conversation(messages1)
    manager1.close()

    # Create second conversation
    manager2 = ConversationManager(db_path=temp_db_path, use_last_conversation=False)
    messages2 = [{"role": "user", "content": "Second conversation"}]
    manager2.update_conversation(messages2)
    manager2.close()

    # Verify we can still access both conversations
    db = TinyDB(temp_db_path)
    table = db.table("conversations")

    assert len(table.all()) == 2

    stored_data1 = table.all()[0]  # First conversation
    stored_data2 = table.all()[1]  # Second conversation
    stored_messages1 = stored_data1["messages"]
    stored_messages2 = stored_data2["messages"]

    assert len(stored_messages1) == 1
    assert stored_messages1[0]["role"] == "user"
    assert stored_messages1[0]["content"] == "First conversation"
    assert "timestamp" in stored_messages1[0]

    assert len(stored_messages2) == 1
    assert stored_messages2[0]["role"] == "user"
    assert stored_messages2[0]["content"] == "Second conversation"
    assert "timestamp" in stored_messages2[0]

    db.close()


def test_conversation_persistence(temp_db_path):
    """Test that conversations persist across manager instances."""

    # Create and populate a conversation
    manager = ConversationManager(db_path=temp_db_path)
    test_messages = [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi there!"}]
    manager.update_conversation(test_messages)

    # Create new manager instance and verify persistence
    manager = ConversationManager(db_path=temp_db_path, use_last_conversation=True)

    assert len(manager.messages) == 2
    assert manager.messages[0]["role"] == "user"
    assert manager.messages[0]["content"] == "Hello"
    assert "timestamp" in manager.messages[0]
    assert manager.messages[1]["role"] == "assistant"
    assert manager.messages[1]["content"] == "Hi there!"
    assert "timestamp" in manager.messages[1]

    manager.close()


def test_complex_message_structure(temp_db_path):
    """Test with complex message structures."""

    manager = ConversationManager(db_path=temp_db_path)

    complex_messages = [
        {"role": "user", "content": "Complex message", "metadata": {"timestamp": "2023-01-01T00:00:00Z"}},
        {
            "role": "assistant",
            "content": "Complex response",
            "tool_calls": [{"id": "call_1", "function": {"name": "test_func"}}],
        },
    ]

    manager.update_conversation(complex_messages)

    # Verify complex structure is preserved and timestamps are added
    updated_messages = manager.messages
    assert len(updated_messages) == 2
    assert updated_messages[0]["role"] == "user"
    assert updated_messages[0]["content"] == "Complex message"
    assert "timestamp" in updated_messages[0]
    assert updated_messages[0]["metadata"] == {"timestamp": "2023-01-01T00:00:00Z"}

    assert updated_messages[1]["role"] == "assistant"
    assert updated_messages[1]["content"] == "Complex response"
    assert "timestamp" in updated_messages[1]
    assert updated_messages[1]["tool_calls"] == [{"id": "call_1", "function": {"name": "test_func"}}]

    manager.close()


def test_timestamp_format(temp_db_path):
    """Test that timestamps are in ISO format."""

    manager = ConversationManager(db_path=temp_db_path)

    messages = [{"role": "user", "content": "Hello"}]
    manager.update_conversation(messages)

    updated_messages = manager.messages
    timestamp = updated_messages[0]["timestamp"]

    # Verify it's a valid ISO format timestamp
    from datetime import datetime

    datetime.fromisoformat(timestamp)  # This will raise an exception if invalid

    manager.close()



from tqdm import tqdm
import logging

logger = logging.getLogger(__name__)

DEFAULT_GENERATE_KWARGS = {"max_new_tokens": 3, "temperature": 0.01, "do_sample": True}

CLASSIFICATION_TASK_PROMPT_TEMPLATE = """
Read the following scientific paper abstract. Based on the content, determine if the paper explicitly refers to or uses a disease modeling technique,
including but not limited to mathematical, statistical, or computational methods used to simulate, analyze, predict, or interpret the dynamics of a disease.
"""

CLASSIFICATION_OUTPUT_PROMPT_TEMPLATE = """
If the abstract describes or references any of these methods or similar approaches, answer "YES".
If the abstract focuses on non-modeling analysis, such as reporting observational data without reference to disease modeling techniques, answer "NO".
Wrap your final answer in square braces, e.g. [YES] or [NO]. Only include [YES] or [NO] in your answer.

Abstract:
{abstract}
"""


class ClassificationPromptTuner:
    def __init__(self, model_client, task_prompt: str = None):
        self.model_client = model_client
        self.initial_task_prompt = task_prompt
        self.tuned_prompt = None

        self.generate_kwargs = DEFAULT_GENERATE_KWARGS.copy()

        # Reasoning models need more generation tokens (for thinking) and a higher temperature for better results
        if model_client.model_id in model_client.THINKING_MODELS:
            self.generate_kwargs.update(
                {
                    "max_new_tokens": 1024,
                    "temperature": 0.6,
                }
            )

    def tune_prompt(self, training_data):
        """
        Executes the training process (prompt auto tuning) for classification using a hill climbing approach to optimize the prompt for 100% accuracy.

        Args:
            data: DataFrame with training data containing 'abstract' and 'is_modeling' columns

        Returns:
            Tuned prompt string
        """

        df_data = training_data.copy()
        df_data["predict_modeling"] = None

        # Use initial prompt if no tuned prompt exists
        if self.tuned_prompt is None:
            self.tuned_prompt = self.initial_task_prompt

        last_prompt = self.tuned_prompt
        iteration = mutation = 0

        # iterate using a basic hill climbing approach until a prompt is found that classifies w/100% accuracy
        while True:
            # Test current prompt
            df_data = self.classify_data(df_data)
            metrics = self.evaluate_results(df_data)

            out = f"results: iteration: {iteration}, mutation: {mutation} - precision: {metrics['precision']:.2f}. recall: {metrics['recall']:.2f}, accuracy: {metrics['accuracy']:.2f}"
            print(out)
            logger.info(out)

            iteration += 1

            # if accuracy has decreased, revert to the last prompt and try again
            if iteration > 1 and metrics["accuracy"] < self.last_metrics["accuracy"]:
                print("reverting to last prompt")
                self.tuned_prompt = last_prompt
                mutation -= 1
                continue

            df_bad = df_data.query("is_modeling != predict_modeling")

            # stop iterating once accuracy is 100%
            if len(df_bad) == 0:
                break

            # Store current state
            last_prompt = self.tuned_prompt
            self.last_metrics = metrics

            # randomly select an incorrect result for mutating the task prompt
            item = df_bad.sample().iloc[0]
            if item.is_modeling:
                mutation_prompt = self._get_positive_mutation_prompt(self.tuned_prompt, item.abstract)
            else:
                mutation_prompt = self._get_negative_mutation_prompt(self.tuned_prompt, item.abstract)

            print("mutating prompt")
            logger.info(f"mutation prompt:\n{mutation_prompt}")

            # generate a new prompt using the mutation prompt
            result = self.model_client.generate(mutation_prompt, self.generate_kwargs)
            logger.info(f"raw prompt mutation results:\n{result}")

            # extract the prompt from the response
            mutated_prompt = result.split("<prompt>")[-1].split("</prompt>")[0].strip()
            logger.info(f"processed prompt mutation results:\n{mutated_prompt}")

            mutation += 1
            self.tuned_prompt = mutated_prompt

        return self.tuned_prompt

    def classify_data(self, classification_prompt, data):
        """
        Execute classification task on the provided data using current prompt.

        Args:
            data: DataFrame with a 'content' column for classification.

        Returns:
            DataFrame with added 'predicted_class' column.
        """

        predicted_class = []
        df = data.copy()

        for i in tqdm(range(len(data)), desc="classifying"):
            row = data.iloc[i]
            prompt = classification_prompt.format(content=row.content)

            result = self.model_client.generate(prompt, self.generate_kwargs)

            logger.info(f"processed classification results: {result}")

            if "[yes]" in result.lower():
                predicted_class.append(True)
            elif "[no]" in result.lower():
                predicted_class.append(False)
            else:
                raise ValueError(f"unexpected classification result: {result}")

            logger.info(f"classification: {predicted_class[-1]}")

        df["predicted_class"] = predicted_class
        return df

    def evaluate_results(self, data):
        """
        Evaluate classification results and return metrics.

        Args:
            data: DataFrame with 'actual_class' and 'predicted_class' columns

        Returns:
            Dictionary with accuracy, precision, and recall metrics
        """

        true_pos = len(data.query("actual_class == True and predicted_class == True"))
        true_neg = len(data.query("actual_class == False and predicted_class == False"))

        accuracy = (true_pos + true_neg) / len(data)
        pos = len(data.query("predicted_class == True"))
        precision = true_pos / pos if pos > 0 else 0
        pos = len(data.query("actual_class == True"))
        recall = true_pos / pos if pos > 0 else 0

        return {"accuracy": accuracy, "precision": precision, "recall": recall}

    def _get_positive_mutation_prompt(self, prompt, abstract):
        """Generate mutation prompt for positive examples."""
        return f"""Given this task prompt:
{prompt}

This abstract should be classified as YES (modeling paper) but was classified as NO:
{abstract}

Please improve the task prompt to better identify modeling papers like this one. Return the improved prompt wrapped in <prompt></prompt> tags."""

    def _get_negative_mutation_prompt(self, prompt, abstract):
        """Generate mutation prompt for negative examples."""
        return f"""Given this task prompt:
{prompt}

This abstract should be classified as NO (non-modeling paper) but was classified as YES:
{abstract}

Please improve the task prompt to better exclude non-modeling papers like this one. Return the improved prompt wrapped in <prompt></prompt> tags."""



import os
import pytest

from aimu.prompts import Prompt, PromptCatalog


@pytest.fixture
def prompt_catalog() -> PromptCatalog:
    catalog = PromptCatalog("test.db")
    yield catalog
    os.remove("test.db")


def test_store_prompt(prompt_catalog):
    model_id = "x"

    prompt1 = Prompt(prompt="test", model_id=model_id, version=1)
    prompt_catalog.store_prompt(prompt1)
    prompt2 = prompt_catalog.retrieve_last(model_id)

    assert prompt1.id == prompt2.id


def test_retrieve_last(prompt_catalog):
    model_id = "x"

    prompt1 = Prompt(prompt="a", model_id=model_id, version=1)
    prompt_catalog.store_prompt(prompt1)

    prompt2 = Prompt(prompt="c", model_id=model_id, version=3)
    prompt_catalog.store_prompt(prompt2)

    prompt3 = Prompt(prompt="b", model_id=model_id, version=2)
    prompt_catalog.store_prompt(prompt3)

    prompt4 = prompt_catalog.retrieve_last(model_id)

    assert prompt2.id == prompt4.id


def test_retrieve_all(prompt_catalog):
    model_id = "x"

    prompt1 = Prompt(prompt="a", model_id=model_id)
    prompt_catalog.store_prompt(prompt1)

    prompt2 = Prompt(prompt="c", model_id=model_id)
    prompt_catalog.store_prompt(prompt2)

    prompts = prompt_catalog.retrieve_all(model_id)

    assert prompt1 in prompts
    assert prompt2 in prompts


def test_delete_all(prompt_catalog):
    model_id = "x"

    prompt1 = Prompt(prompt="a", model_id=model_id)
    prompt_catalog.store_prompt(prompt1)

    prompt2 = Prompt(prompt="c", model_id=model_id)
    prompt_catalog.store_prompt(prompt2)

    prompts = prompt_catalog.retrieve_all(model_id)

    assert prompt1 in prompts
    assert prompt2 in prompts

    rows_deleted = prompt_catalog.delete_all(model_id)

    assert rows_deleted == 2

    prompts = prompt_catalog.retrieve_all(model_id)

    assert len(prompts) == 0


def test_retrieve_model_ids(prompt_catalog):
    model_id = "x"

    prompt1 = Prompt(prompt="a", model_id=model_id)
    prompt_catalog.store_prompt(prompt1)

    model_ids = prompt_catalog.retrieve_model_ids()

    assert len(model_ids) == 1
    assert model_ids[0] == "x"



from ..base_client import Model, ModelClient

import torch
from transformers import AutoTokenizer
from transformers import AutoModelForCausalLM
from transformers.utils import logging as log
from transformers import TextIteratorStreamer
import gc
import pprint
import logging
from typing import Iterator, Optional, Any
import json
import itertools

logger = logging.getLogger(__name__)
log.set_verbosity_error()


class HuggingFaceModel(Model):
    GPT_OSS_20B = "openai/gpt-oss-20b"

    LLAMA_3_1_8B = "meta-llama/Meta-Llama-3.1-8B-Instruct"
    LLAMA_3_2_3B = "unsloth/Llama-3.2-3B-Instruct"

    DEEPSEEK_R1_8B = "deepseek-ai/DeepSeek-R1-Distill-Llama-8B"

    GEMMA_3_12B = "google/gemma-3-12b-it"

    PHI_4_14B = "microsoft/phi-4"
    PHI_4_MINI_3_8B = "microsoft/Phi-4-mini-instruct"

    MISTRAL_7B = "mistralai/Mistral-7B-Instruct-v0.3"
    MISTRAL_NEMO_12B = "mistralai/Mistral-Nemo-Instruct-2407"
    MISTRAL_SMALL_3_2_24B = "mistralai/Mistral-Small-3.2-24B-Instruct-2506"

    QWEN_3_8B = "Qwen/Qwen3-8B"

    SMOLLM3_3B = "HuggingFaceTB/SmolLM3-3B"


class HuggingFaceClient(ModelClient):
    MODELS = HuggingFaceModel

    TOOL_MODELS = [
        # MODELS.GPT_OSS_20B,
        MODELS.QWEN_3_8B,
        MODELS.MISTRAL_7B,
        MODELS.MISTRAL_NEMO_12B,
        MODELS.MISTRAL_SMALL_3_2_24B,
        # MODELS.LLAMA_3_1_8B,
        # MODELS.LLAMA_3_2_3B,
        # MODELS.SMOLLM3_3B,
    ]

    THINKING_MODELS = [
        MODELS.GPT_OSS_20B,
        MODELS.DEEPSEEK_R1_8B,
        MODELS.QWEN_3_8B,
        MODELS.SMOLLM3_3B,
    ]

    DEFAULT_MODEL_KWARGS = {
        "device_map": "auto",
        "torch_dtype": "auto",
    }

    DEFAULT_GENERATE_KWARGS = {
        "max_new_tokens": 1024,
        "temperature": 0.1,
        "top_p": 0.95,
        "top_k": 50,
        "repetition_penalty": 1.1,
        "do_sample": True,
        "num_beams": 1,
    }

    DEFAULT_MODEL_GENERATE_KWARGS = {
        MODELS.GPT_OSS_20B.value: {
            "temperature": 1.0,
            "top_p": 1.0,
            "top_k": 0,
        },
        MODELS.QWEN_3_8B.value: {
            ## thinking
            "temperature": 0.6,
            "top_p": 0.95,
            "top_k": 20,
            "min_p": 0,
            ## TODO support non-thinking defaults
            # "temperature": 0.7,
            # "top_p": 0.8,
        },
    }

    def __init__(
        self,
        model: HuggingFaceModel,
        model_kwargs: dict = DEFAULT_MODEL_KWARGS.copy(),
        system_message: Optional[str] = None,
    ):
        super().__init__(model, model_kwargs, system_message)

        self.hf_tokenizer = AutoTokenizer.from_pretrained(model.value, attn_implementation="eager")
        self.hf_model = AutoModelForCausalLM.from_pretrained(model.value, **model_kwargs)

        self.default_generate_kwargs = self.DEFAULT_GENERATE_KWARGS.copy()

        if self.model in self.DEFAULT_MODEL_GENERATE_KWARGS:
            self.default_generate_kwargs.update(self.DEFAULT_MODEL_GENERATE_KWARGS[self.model.value])

    def __del__(self):
        del self.hf_model
        del self.hf_tokenizer

        gc.collect()

        if torch.cuda.is_available():
            logger.info("Emptying CUDA cache")
            torch.cuda.empty_cache()
        elif torch.backends.mps.is_available():
            logger.info("Emptying MPS cache")
            torch.mps.empty_cache()

    def _update_generate_kwargs(self, generate_kwargs: Optional[dict[str, Any]] = None) -> dict[str, None]:
        if not generate_kwargs:
            generate_kwargs = self.default_generate_kwargs

        if "max_tokens" in generate_kwargs:
            generate_kwargs["max_new_tokens"] = generate_kwargs.pop("max_tokens")

        if "repeat_penalty" in generate_kwargs:
            generate_kwargs["repetition_penalty"] = generate_kwargs.pop("repeat_penalty")

        return generate_kwargs

    def _generate(
        self,
        messages: list[dict],
        generate_kwargs: dict[str, Any],
        tools: Optional[list[dict]] = None,
        streamer: Optional[TextIteratorStreamer] = None,
    ) -> tuple[str, Iterator[str]]:
        if (tools and len(tools) == 0) or self.model not in self.TOOL_MODELS:
            ## some models (e.g. Llama 3.1) misbehave is tools is present, event if None
            text = self.hf_tokenizer.apply_chat_template(
                messages,
                add_generation_prompt=True,
                return_tensors="pt",
                tokenize=False,
                enable_thinking=True,
            )
        else:
            text = self.hf_tokenizer.apply_chat_template(
                messages,
                add_generation_prompt=True,
                return_tensors="pt",
                tokenize=False,
                enable_thinking=True,
                tools=tools,
                xml_tools=tools,
            )

        self.last_thinking = ""

        model_inputs = self.hf_tokenizer([text], return_tensors="pt").to(self.hf_model.device)
        generated_ids = self.hf_model.generate(**model_inputs, **generate_kwargs, streamer=streamer)

        if streamer is None:
            output_ids = generated_ids[0][len(model_inputs.input_ids[0]) :]
            response = self.hf_tokenizer.decode(output_ids, skip_special_tokens=True)

            if self.model in self.THINKING_MODELS and response.startswith("<think>"):
                self.last_thinking = response[len("<think>") : response.index("</think>")].strip()
                response = response[response.index("</think>") + len("</think>") :].strip()

            return response, iter([])

        next(streamer)
        a, b = itertools.tee(streamer)

        response_part = next(a)

        if self.model in self.THINKING_MODELS and response_part.startswith("<think>"):
            while True:
                response_part = next(a)
                next(b)

                if "</think>" in response_part:
                    next(b)
                    break

                self.last_thinking += response_part

            self.last_thinking = self.last_thinking.strip()

        return "", b

    def generate(self, prompt: str, generate_kwargs: Optional[dict[str, Any]] = None) -> str:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        messages = [
            {"role": "user", "content": prompt},
        ]

        response, _ = self._generate(messages, generate_kwargs)

        return response

    def generate_streamed(self, prompt: str, generate_kwargs: Optional[dict[str, Any]] = None) -> Iterator[str]:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        messages = [
            {"role": "user", "content": prompt},
        ]

        streamer = TextIteratorStreamer(self.hf_tokenizer, skip_prompt=True)

        _, it = self._generate(messages, generate_kwargs, streamer=streamer)

        return it

    def chat(self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True) -> str:
        generate_kwargs, tools = self._chat_setup(user_message, generate_kwargs, use_tools)

        response, _ = self._generate(self.messages, generate_kwargs, tools)

        if self.model in [self.MODELS.QWEN_3_8B, self.MODELS.SMOLLM3_3B, self.MODELS.GPT_OSS_20B]:
            if response.startswith("<tool_call>"):
                tool_calls = []

                while "<tool_call>" in response:
                    tool_call_start = response.index("<tool_call>") + len("<tool_call>")
                    tool_call_end = response.index("</tool_call>")
                    tool_call = response[tool_call_start:tool_call_end].strip()
                    tool_calls.append(json.loads(tool_call))

                    response = response[tool_call_end + len("</tool_call>") :].strip()

                self._handle_tool_calls(tool_calls, tools)

                # assign thinking to the tool call (the second to last message, before the tool response)
                # TODO determine what to do when there are multuple tool calls
                if self.last_thinking:
                    self.messages[-2]["thinking"] = self.last_thinking

                response, _ = self._generate(self.messages, generate_kwargs, tools)
        elif self.model in [self.MODELS.MISTRAL_SMALL_3_2_24B]:
            if "[TOOL_CALLS]" in response:
                self._handle_tool_calls(json.loads(response), tools)
                response, _ = self._generate(self.messages, generate_kwargs, tools)
        elif self.model in [self.MODELS.LLAMA_3_1_8B, self.MODELS.LLAMA_3_2_3B]:
            if response.startswith('{"name":'):
                self._handle_tool_calls([json.loads(response)], tools)
                response, _ = self._generate(self.messages, generate_kwargs, tools)
        elif self.model in [self.MODELS.MISTRAL_7B, self.MODELS.MISTRAL_NEMO_12B]:
            if response.startswith('[{"name":'):
                self._handle_tool_calls([json.loads(response)[0]], tools)
                response, _ = self._generate(self.messages, generate_kwargs, tools)

        self.messages.append({"role": "assistant", "content": response})

        if self.last_thinking:
            self.messages[-1]["thinking"] = self.last_thinking

        return response

    def chat_streamed(
        self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True
    ) -> Iterator[str]:
        generate_kwargs, tools = self._chat_setup(user_message, generate_kwargs, use_tools)

        streamer = TextIteratorStreamer(self.hf_tokenizer, skip_prompt=True, skip_special_tokens=False)

        _, it = self._generate(self.messages, generate_kwargs, tools, streamer=streamer)

        response_part = next(it)
        content = ""

        if self.model in [self.MODELS.QWEN_3_8B, self.MODELS.SMOLLM3_3B]:
            if "<tool_call>" in response_part:
                response = response_part
                for response_part in it:
                    response += response_part

                tool_calls = []
                while "<tool_call>" in response:
                    tool_call_start = response.index("<tool_call>") + len("<tool_call>")
                    tool_call_end = response.index("</tool_call>")
                    tool_call = response[tool_call_start:tool_call_end].strip()

                    tool_calls.append(json.loads(tool_call))

                    response = response[tool_call_end + len("</tool_call>") :].strip()

                self._handle_tool_calls(tool_calls, tools)

                if self.last_thinking:
                    self.messages[-2]["thinking"] = self.last_thinking

                _, it = self._generate(self.messages, generate_kwargs, tools, streamer=streamer)
            else:
                content += response_part
        elif self.model in [self.MODELS.MISTRAL_SMALL_3_2_24B]:
            if "[TOOL_CALLS]" in response_part:
                response = response_part
                for response_part in it:
                    response += response_part

                tool_calls_start = response.index("[TOOL_CALLS]") + len("[TOOL_CALLS]")
                tool_calls = response[tool_calls_start:].strip()

                self._handle_tool_calls(json.loads(tool_calls), tools)
                _, it = self._generate(self.messages, generate_kwargs, tools, streamer=streamer)
            else:
                content += response_part
        elif self.model in [self.MODELS.LLAMA_3_1_8B, self.MODELS.LLAMA_3_2_3B]:
            if '{"name":' in response_part:
                response = response_part
                for response_part in it:
                    response += response_part

                self._handle_tool_calls([json.loads(response)], tools)
                _, it = self._generate(self.messages, generate_kwargs, tools)
        elif self.model in [self.MODELS.MISTRAL_7B, self.MODELS.MISTRAL_NEMO_12B]:
            if '[{"name":' in response_part:
                response = response_part
                for response_part in it:
                    response += response_part

                self._handle_tool_calls([json.loads(response)[0]], tools)
                _, it = self._generate(self.messages, generate_kwargs, tools)

        while True:
            yield response_part

            try:
                response_part = next(it)
                content += response_part
                yield response_part
            except StopIteration:
                break

        self.messages.append({"role": "assistant", "content": content})

        if self.last_thinking:
            self.messages[-1]["thinking"] = self.last_thinking

        return

    def print_model_info(self):
        print(f"model : size : {self.hf_model.get_memory_footprint() // 1024**2} MB")

        try:
            print(f"model : is quantized : {self.hf_model.is_quantized}")
            print(f"model : quantization method : {self.hf_model.quantization_method}")
        except AttributeError:
            print("model : is quantized : False")
            pass

        try:
            print(f"model : 8-bit quantized : {self.hf_model.is_loaded_in_8bit}")
        except AttributeError:
            pass

        try:
            print(f"model : 4-bit quantized : {self.hf_model.is_loaded_in_4bit}")
        except AttributeError:
            pass

        param = next(self.hf_model.parameters())
        print(f"model : on GPU (CUDA) : {param.is_cuda}")
        print(f"model : on GPU (MPS) : {param.is_mps}")

    def print_device_map(self):
        pp = pprint.PrettyPrinter(indent=2)
        pp.pprint(self.hf_model.hf_device_map)



from sqlalchemy import (
    String,
    Integer,
    PickleType,
    Column,
    create_engine,
    ForeignKey,
    desc,
)
from sqlalchemy.orm import sessionmaker, declarative_base

Base = declarative_base()


class Prompt(Base):
    __tablename__ = "prompts"
    id = Column(Integer, primary_key=True)
    parent_id = Column(Integer, ForeignKey("prompts.id"))
    prompt = Column(String, nullable=False)
    model_id = Column(String)
    version = Column(Integer)
    mutation_prompt = Column(String)
    reasoning_prompt = Column(String)
    metrics = Column(PickleType)


class PromptCatalog:
    def __init__(self, db_path: str):
        self.engine = create_engine(f"sqlite:///{db_path}")
        Base.metadata.create_all(self.engine)

        Session = sessionmaker(bind=self.engine)
        self.session = Session()

    def __del__(self):
        self.session.close()

    def store_prompt(self, prompt: Prompt):
        try:
            self.session.add(prompt)
        except:
            self.session.rollback()
            raise
        else:
            self.session.commit()

    def retrieve_last(self, model_id: str) -> Prompt:
        return self.session.query(Prompt).filter(Prompt.model_id == model_id).order_by(desc(Prompt.version)).first()

    def retrieve_all(self, model_id: str) -> list[Prompt]:
        return self.session.query(Prompt).filter(Prompt.model_id == model_id).order_by(desc(Prompt.version)).all()

    def delete_all(self, model_id: str) -> int:
        rows_deleted = 0

        try:
            rows_deleted = self.session.query(Prompt).filter(Prompt.model_id == model_id).delete()
        except:
            self.session.rollback()
            raise
        else:
            self.session.commit()

        return rows_deleted

    def retrieve_model_ids(self):
        return [x.model_id for x in self.session.query(Prompt.model_id).distinct()]



from typing import Any, Iterator, Optional
from ..base_client import Model, ModelClient

import aisuite
import logging
from dotenv import load_dotenv
import json

logger = logging.getLogger(__name__)


class AisuiteModel(Model):
    GPT_4O_MINI = "openai:gpt-4o-mini"
    GPT_4O = "openai:gpt-4o"
    ## TODO: re-enable GPT-5 models once issue with "max_completion_tokens" is resolved
    # GPT_5_NANO = "openai:gpt-5-nano"
    # GPT_5_MINI = "openai:gpt-5-mini"
    # GPT_5 = "openai:gpt-5"


class AisuiteClient(ModelClient):
    MODELS = AisuiteModel
    TOOL_MODELS = MODELS
    THINKING_MODELS = []

    DEFAULT_GENERATE_KWARGS = {
        "max_new_tokens": 1024,
        "temperature": 0.1,
        "top_p": 0.95,
        "top_k": 50,
    }

    def __init__(
        self,
        model: AisuiteModel,
        model_kwargs: Optional[dict] = None,
        system_message: Optional[str] = None,
    ):
        super().__init__(model, model_kwargs, system_message)
        self.default_generate_kwargs = self.DEFAULT_GENERATE_KWARGS.copy()

        load_dotenv()

        self.ai_client = aisuite.Client()

    def _update_generate_kwargs(self, generate_kwargs: Optional[dict[str, Any]] = None) -> dict[str, None]:
        if not generate_kwargs:
            generate_kwargs = self.default_generate_kwargs
        else:
            generate_kwargs = {**self.default_generate_kwargs, **generate_kwargs}

        return generate_kwargs

    def generate(self, prompt: str, generate_kwargs: Optional[dict[str, Any]] = None) -> str:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        messages = [{"role": "user", "content": prompt}]

        response = self.ai_client.chat.completions.create(
            self.model.value,
            messages,
            temperature=generate_kwargs["temperature"],
            max_tokens=generate_kwargs["max_new_tokens"],
        )

        return response.choices[0].message.content

    def generate_streamed(self, prompt: str, generate_kwargs: Optional[dict[str, Any]] = None) -> Iterator[str]:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        messages = [{"role": "user", "content": prompt}]

        response = self.ai_client.chat.completions.create(
            self.model.value,
            messages,
            stream=True,
            temperature=generate_kwargs["temperature"],
            max_tokens=generate_kwargs["max_new_tokens"],
        )

        message = {"role": next(response).choices[0].delta.role}
        content = ""

        while response_part := next(response):
            if response_part.choices[0].finish_reason is not None:
                break

            content += response_part.choices[0].delta.content
            yield response_part.choices[0].delta.content

        message["content"] = content

    def chat(self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True) -> str:
        generate_kwargs, tools = self._chat_setup(user_message, generate_kwargs, use_tools)

        response = self.ai_client.chat.completions.create(
            self.model.value,
            self.messages,
            tools=tools,
            temperature=generate_kwargs["temperature"],
            max_tokens=generate_kwargs["max_new_tokens"],
        )

        if response.choices[0].message.tool_calls:
            tool_calls = []
            for tool_call in response.choices[0].message.tool_calls:
                tool_calls.append(
                    {
                        "name": tool_call.function.name,
                        "arguments": json.loads(tool_call.function.arguments),
                    }
                )

            self._handle_tool_calls(tool_calls, tools)

            ## tool calls for aisuite require the arguments to be a json string
            # TODO: find a better way. this makes messages inconsistent between Aisuite and other model clients
            self.messages[-2]["tool_calls"][0]["function"]["arguments"] = json.dumps(
                self.messages[-2]["tool_calls"][0]["function"]["arguments"]
            )

            response = self.ai_client.chat.completions.create(
                self.model.value,
                self.messages,
                tools=tools,
                temperature=generate_kwargs["temperature"],
                max_tokens=generate_kwargs["max_new_tokens"],
            )

        self.messages.append({"role": "assistant", "content": response.choices[0].message.content})

        return response.choices[0].message.content

    def chat_streamed(
        self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True
    ) -> Iterator[str]:
        generate_kwargs, tools = self._chat_setup(user_message, generate_kwargs, use_tools)

        response = self.ai_client.chat.completions.create(
            self.model.value,
            self.messages,
            stream=True,
            tools=tools,
            temperature=generate_kwargs["temperature"],
            max_tokens=generate_kwargs["max_new_tokens"],
        )

        response_part = next(response)

        if response_part.choices[0].delta.tool_calls:
            function_name = response_part.choices[0].delta.tool_calls[0].function.name

            function_arguments = ""
            while response_part := next(response):
                if response_part.choices[0].finish_reason is not None:
                    break
                
                function_arguments += response_part.choices[0].delta.tool_calls[0].function.arguments

            tool_calls = [] ## TODO: handle multiple tool calls
            tool_calls.append(
                {
                    "name": function_name,
                    "arguments": json.loads(function_arguments),
                }
            )

            self._handle_tool_calls(tool_calls, tools)

            ## tool calls for aisuite require the arguments to be a json string
            # TODO: find a better way. this makes messages inconsistent between Aisuite and other model clients
            self.messages[-2]["tool_calls"][0]["function"]["arguments"] = json.dumps(
                self.messages[-2]["tool_calls"][0]["function"]["arguments"]
            )

            response = self.ai_client.chat.completions.create(
                self.model.value,
                self.messages,
                stream=True,
                tools=tools,
                temperature=generate_kwargs["temperature"],
                max_tokens=generate_kwargs["max_new_tokens"],
            )

        message = {"role": next(response).choices[0].delta.role}
        content = ""

        while response_part := next(response):
            if response_part.choices[0].finish_reason is not None:
                break

            content += response_part.choices[0].delta.content
            yield response_part.choices[0].delta.content

        message["content"] = content
        self.messages.append(message)



from fastmcp import FastMCP

mcp = FastMCP("AIMU Tools")


@mcp.tool()
def echo(echo_string: str) -> str:
    """Returns echo_string."""
    return echo_string


@mcp.tool()
def get_current_data_and_time() -> str:
    """Returns the current date and time."""
    from datetime import datetime

    return datetime.now().isoformat()


if __name__ == "__main__":
    mcp.run()



from fastmcp import FastMCP
from fastmcp import Client

import asyncio
from typing import Optional


class MCPClient:
    """A MCP client that wraps the FastMCP client for easier tool calls and management."""

    def __init__(self, config: Optional[dict] = None, server: Optional[FastMCP] = None, file: Optional[str] = None):
        self.loop = asyncio.new_event_loop()

        if config:
            self.mcp_config = config
        if file:
            self.mcp_config = file
        if server:
            self.mcp_config = server

        self.client = Client(self.mcp_config)
        self.loop.run_until_complete(self.client.__aenter__())

    def __del__(self):
        self.loop.run_until_complete(self.client.__aexit__(None, None, None))
        if self.loop.is_running():
            self.loop.stop()

    async def _call_tool(self, tool_name: str, params: dict):
        return await self.client.call_tool(tool_name, params)

    def call_tool(self, tool_name: str, params: dict):
        return self.loop.run_until_complete(self._call_tool(tool_name, params))

    async def _list_tools(self):
        return await self.client.list_tools()

    def list_tools(self):
        return self.loop.run_until_complete(self._list_tools())

    def get_tools(self) -> list[dict]:
        tools = []

        for tool in self.list_tools():
            tools.append(
                {
                    "type": "function",
                    "function": {"name": tool.name, "description": tool.description, "parameters": tool.inputSchema},
                }
            )

        return tools



from aimu.models.ollama.ollama_client import OllamaClient
from aimu.prompts import ClassificationPromptTuner

import pytest
import pandas as pd


@pytest.fixture(scope="session")
def model_client():
    client = OllamaClient(OllamaClient.MODELS.LLAMA_3_2_3B)
    yield client
    del client


@pytest.fixture(scope="session")
def thinking_model_client():
    client = OllamaClient(OllamaClient.MODELS.QWEN_3_8B)
    yield client
    del client


def test_classify_data(model_client):
    """Test classification with a standard model."""

    prompt_tuner = ClassificationPromptTuner(model_client=model_client)

    classification_prompt = "Is the following content related to large language models?\nWrap your final answer in square braces, e.g. [YES] or [NO]. Only include [YES] or [NO] in your answer.\nContent: {content}"

    df = pd.DataFrame(
        {
            "content": [
                "Large language models are a type of AI.",
                "This is a random sentence.",
                "LLMs can generate text based on prompts.",
                "This sentence is not related to AI.",
            ]
        }
    )

    df = prompt_tuner.classify_data(classification_prompt, df)

    assert "predicted_class" in df.columns
    assert len(df.predicted_class) == 4
    assert df.predicted_class.value_counts().get(True) + df.predicted_class.value_counts().get(False) == 4


def test_classify_data_thinking_model(thinking_model_client):
    """Test classification with a thinking model."""

    test_classify_data(thinking_model_client)


def test_evaluate_results(model_client):
    """Test the evaluation of classification results."""

    prompt_tuner = ClassificationPromptTuner(model_client=model_client)

    df = pd.DataFrame(
        {
            "actual_class": [True, False, True, False],
            "predicted_class": [True, False, True, False],
        }
    )

    metrics = prompt_tuner.evaluate_results(df)

    assert metrics["accuracy"] == 1.0, "Accuracy should be 100%."
    assert metrics["precision"] == 1.0, "Precision should be 100%."
    assert metrics["recall"] == 1.0, "Recall should be 100%."

    df = pd.DataFrame(
        {
            "actual_class": [True, True, False, False],
            "predicted_class": [True, False, True, False],
        }
    )

    metrics = prompt_tuner.evaluate_results(df)

    assert metrics["accuracy"] == 0.5, "Accuracy should be 50%."
    assert metrics["precision"] == 0.5, "Precision should be 50%."
    assert metrics["recall"] == 0.5, "Recall should be 50%."



# Always available
from .base_client import ModelClient

# Optional imports with graceful fallbacks
try:
    from .hf import HuggingFaceClient

    HAS_HF = True
except ImportError:
    HAS_HF = False
    HuggingFaceClient = None

try:
    from .ollama import OllamaClient

    HAS_OLLAMA = True
except ImportError:
    HAS_OLLAMA = False
    OllamaClient = None

try:
    from .aisuite import AisuiteClient

    HAS_AISUITE = True
except ImportError:
    HAS_AISUITE = False
    AisuiteClient = None

# Expose what's available
__all__ = ["ModelClient"]
if HAS_HF:
    __all__.append("HuggingFaceClient")
if HAS_OLLAMA:
    __all__.append("OllamaClient")
if HAS_AISUITE:
    __all__.append("AisuiteClient")



import logging
from enum import Enum
from typing import Optional, Iterator, Any
from abc import ABC, abstractmethod
import random
import string

logger = logging.getLogger(__name__)


class Model(Enum):
    pass


class ModelClient(ABC):
    MODELS = Model
    TOOL_MODELS = []
    THINKING_MODELS = []

    @abstractmethod
    def __init__(self, model: Model, model_kwargs: Optional[dict] = None, system_message: Optional[str] = None):
        self.model = model
        self.model_kwargs = model_kwargs
        self._system_message = system_message
        self.default_generate_kwargs = {}
        self.messages = []
        self.mcp_client = None
        self.last_thinking = ""

    @property
    def system_message(self):
        return self._system_message

    @system_message.setter
    def system_message(self, message: str):
        self._system_message = message

        # TODO: add support for models that don't have a system message
        # TODO: move system message handling out of the sub classes
        if self.messages:
            if self.messages[0]["role"] == "system":
                self.messages[0]["content"] = message
            else:
                self.messages.insert(0, {"role": "system", "content": message})

    @abstractmethod
    def generate(self, prompt: str, generate_kwargs: Optional[dict[str, Any]] = None) -> str:
        pass

    @abstractmethod
    def generate_streamed(self, prompt: str, generate_kwargs: Optional[dict[str, Any]] = None) -> Iterator[str]:
        pass

    @abstractmethod
    def chat(self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True) -> str:
        pass

    @abstractmethod
    def chat_streamed(
        self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True
    ) -> Iterator[str]:
        pass

    @abstractmethod
    def _update_generate_kwargs(self, generate_kwargs: Optional[dict[str, Any]] = None) -> dict:
        pass

    def _chat_setup(
        self, user_message: str, generate_kwargs: Optional[dict[str, Any]] = None, use_tools: bool = True
    ) -> tuple[dict[str, Any], list[dict[str, Any]]]:
        generate_kwargs = self._update_generate_kwargs(generate_kwargs)

        # Add the system message if we're processing the first user message and system_message is set
        if len(self.messages) == 0 and self.system_message:
            self.messages.append({"role": "system", "content": self.system_message})

        # Add the user message
        self.messages.append({"role": "user", "content": user_message})

        tools = []
        if self.model in self.TOOL_MODELS and use_tools and self.mcp_client:
            tools = self.mcp_client.get_tools()

        return generate_kwargs, tools

    def _handle_tool_calls(self, tool_calls: list[dict], tools: list) -> None:
        message = {"role": "assistant", "tool_calls": []}
        self.messages.append(message)

        for tool_call in tool_calls:
            # llama 3.1 uses 'parameters' instead of 'arguments'
            if "arguments" not in tool_call and "parameters" in tool_call:
                tool_call["arguments"] = tool_call.pop("parameters")

            id = "".join(random.choices(string.ascii_letters + string.digits, k=9))

            message["tool_calls"].append(
                {
                    "type": "function",
                    "function": {"name": tool_call["name"], "arguments": tool_call["arguments"]},
                    "id": id,
                }
            )

            for tool in tools:
                # If the tool is a callable python function, call it directly. Otherwise, use the MCP client to call the tool.
                if hasattr(tool, "__call__") and tool.__name__ == tool_call["name"]:
                    tool_response = tool(**tool_call["arguments"])

                    self.messages.append(
                        {"role": "tool", "name": tool_call["name"], "content": str(tool_response), "tool_call_id": id}
                    )

                    break
                elif tool["type"] == "function" and tool["function"]["name"] == tool_call["name"]:
                    if self.mcp_client is None:
                        raise ValueError(
                            "MCP client not initialized. Please initialize and assign an MCP client before using MCP tools."
                        )

                    tool_response = self.mcp_client.call_tool(tool["function"]["name"], tool_call["arguments"])

                    content = ""
                    if len(tool_response.content) > 0:
                        # TODO: handle different tool response types, errors, and multiple responses
                        if tool_response.content[0].type != "text":
                            raise ValueError(
                                f"Tool response type {tool_response.content[0].type} not supported. Supported types: text"
                            )

                        content = tool_response.content[0].text

                    self.messages.append(
                        {"role": "tool", "name": tool["function"]["name"], "content": content, "tool_call_id": id}
                    )

                    break



"""
Model client tests that can run with different client types.

Usage:
- Run all tests with all models: pytest tests/test_models.py
- Run tests with only Ollama models: pytest tests/test_models.py --client=ollama
- Run tests with only HuggingFace models: pytest tests/test_models.py --client=hf
- Run tests with only Aisuite models: pytest tests/test_models.py --client=aisuite
"""

import pytest
from typing import Iterable
from fastmcp import FastMCP
import time

from aimu.models import ModelClient, HuggingFaceClient, OllamaClient, AisuiteClient
from aimu.tools.client import MCPClient


def pytest_generate_tests(metafunc):
    """Generate tests based on command line options."""

    if "model_client" in metafunc.fixturenames:
        client_type = metafunc.config.getoption("--client", "all").lower()

        if client_type == "ollama":
            test_models = OllamaClient.MODELS
        elif client_type in ["hf", "huggingface"]:
            # test_models = HuggingFaceClient.MODELS
            test_models = [
                model for model in HuggingFaceClient.MODELS if model != HuggingFaceClient.MODELS.MISTRAL_SMALL_3_2_24B
            ]
            # TODO implement workaround for Mistral Small 3.2 24B not supporting AutoTokenizer
        elif client_type == "aisuite":
            test_models = AisuiteClient.MODELS
        else:
            test_models = list(OllamaClient.MODELS) + list(HuggingFaceClient.MODELS)
            # TODO: add AisuiteClient.MODELS once available

        metafunc.parametrize("model_client", test_models, indirect=True, scope="session")


@pytest.fixture(scope="session")
def model_client(request) -> Iterable[ModelClient]:
    """Create model client based on the model parameter."""

    model = request.param

    if model in OllamaClient.MODELS:
        time.sleep(2)  # give Ollama some time to free memory from the prior model
        client = OllamaClient(model, system_message="You are a helpful assistant.", model_keep_alive_seconds=2)
    elif model in HuggingFaceClient.MODELS:
        client = HuggingFaceClient(model, system_message="You are a helpful assistant.")
    elif model in AisuiteClient.MODELS:
        client = AisuiteClient(model, system_message="You are a helpful assistant.")
    else:
        raise ValueError(f"Unknown model: {model}")

    yield client

    del client


def test_class_properties(model_client):
    """Test that the model client is a subclass of ModelClient."""

    assert issubclass(model_client.__class__, ModelClient)
    assert isinstance(model_client, ModelClient)

    assert model_client.default_generate_kwargs is not None


def test_generate(model_client):
    """Test that the model can generate a response to a prompt."""

    prompt = "What is the capital of France?"
    response = model_client.generate(prompt)

    assert isinstance(response, str)
    assert "Paris" in response


def test_generate_streamed(model_client):
    """Test that the model can generate a response to a prompt in a streamed manner."""

    prompt = "What is the capital of France?"
    response = model_client.generate_streamed(prompt)

    assert isinstance(response, Iterable)
    content = next(response)  # type: ignore
    assert isinstance(content, str)

    for response_part in response:
        content += response_part

    assert "Paris" in content
    assert isinstance(content, str)


def test_generate_with_parameters(model_client):
    """Test that the model can generate a response to a prompt with specific generation parameters."""

    prompt = "What is the capital of France?"
    response = model_client.generate(prompt, generate_kwargs={"max_tokens": 1000})
    # TODO validate the response length in tokens

    assert isinstance(response, str)
    assert "Paris" in response


def test_chat(model_client):
    """Test that the model can chat with a user, maintaining a conversation history."""

    model_client.messages = []

    response = model_client.chat("What is the capital of France?")

    assert "Paris" in response
    assert len(model_client.messages) == 3  # system (auto-added), user, assistant


def test_chat_streamed(model_client):
    """Test that the model can chat with a user in a streamed manner, maintaining a conversation history."""

    model_client.messages = []

    response = model_client.chat_streamed("What is the capital of France?")

    assert isinstance(response, Iterable)
    content = next(response)  # type: ignore
    assert isinstance(content, str)

    for response_part in response:
        content += response_part

    assert "Paris" in content
    assert len(model_client.messages) == 3  # system (auto-added), user, assistant


def test_chat_with_tools(model_client):
    """Test that the model can use tools to answer questions from the user."""

    # ensure the chat history is reset and set system message for tools
    model_client.messages = []
    model_client.system_message = "You are a helpful assistant that uses tools to answer questions from the user."

    mcp = FastMCP("Test Tools")

    @mcp.tool()
    def temperature(location: str) -> float:
        """Return the current temperature in Celsius for a given location."""
        return 27.0

    mcp_client = MCPClient(server=mcp)
    model_client.mcp_client = mcp_client

    response = model_client.chat("What is the temperature in Paris?")

    # If the model does not support tools, we shouldn't see tools being used
    if model_client.model not in model_client.TOOL_MODELS:
        assert len(model_client.messages) == 3  # system (auto-added), user, assistant
        return

    assert len(model_client.messages) == 5  # system (auto-added), user, tool call, tool response, assistant
    assert model_client.messages[-2]["role"] == "tool"  # second to last message should be tool response
    assert "27" in response

    # If the model supports thinking, we should have a thinking messages in the last message and in the tool call
    if model_client.model in model_client.THINKING_MODELS:
        # assert "thinking" in model_client.messages[-1] ## GPT OSS does not seem to add thinking to the last message
        assert "thinking" in model_client.messages[-3]


def test_chat_streamed_with_tools(model_client):
    """Test that the model can use tools to answer questions from the user in a streamed manner."""

    # ensure the chat history is reset and set system message for tools
    model_client.messages = []
    model_client.system_message = "You are a helpful assistant that uses tools to answer questions from the user."

    mcp = FastMCP("Test Tools")

    @mcp.tool()
    def temperature(location: str) -> float:
        """Return the current temperature in Celsius for a given location."""
        return 27.0

    mcp_client = MCPClient(server=mcp)
    model_client.mcp_client = mcp_client

    response = model_client.chat_streamed("What is the temperature in Paris?")

    content = ""
    for response_part in response:
        content += response_part

    # If the model does not support tools, we shouldn't see tools being used
    if model_client.model not in model_client.TOOL_MODELS:
        assert len(model_client.messages) == 3  # system (auto-added), user, assistant
        return

    assert len(model_client.messages) == 5  # system (auto-added), user, tool call, tool response, assistant
    assert model_client.messages[-2]["role"] == "tool"  # second to last message should be tool response
    assert "27" in content

    if model_client.model in model_client.THINKING_MODELS:
        # If the model supports thinking, we should have a thinking messages in the last message and in the tool call
        # assert "thinking" in model_client.messages[-1] ## GPT OSS does not seem to add thinking to the last message
        assert "thinking" in model_client.messages[-3]
